#include "RenderAPI.h"

//--------------------------------------------
// VertexFormat
//--------------------------------------------
VertexFormat::VertexFormat(u32 vertexFlags) {
	Set(vertexFlags);
}

void VertexFormat::Set(u32 vertexFlags) {
	mVertexFlags = vertexFlags;
	mNumElements = 0;
	size_t offset = 0;
	if (vertexFlags & VU_POS) {
		mElements[mNumElements] = VertexElement(offset, VU_POS, VT_FLOAT3);
		offset += (3 * sizeof(float));
		mNumElements++;
	}
	if (vertexFlags & VU_NRM) {
		mElements[mNumElements] = VertexElement(offset, VU_NRM, VT_FLOAT3);
		offset += (3 * sizeof(float));
		mNumElements++;
	}
	if (vertexFlags & VU_TAN) {
		mElements[mNumElements] = VertexElement(offset, VU_TAN, VT_FLOAT4);
		offset += (4 * sizeof(float));
		mNumElements++;
	}
	if (vertexFlags & VU_COL) {
		mElements[mNumElements] = VertexElement(offset, VU_COL, VT_COLOR4);
		offset += (sizeof(u32));
		mNumElements++;
	}
	if (vertexFlags & VU_UV0) {
		mElements[mNumElements] = VertexElement(offset, VU_UV0, VT_FLOAT2);
		offset += (2 * sizeof(float));
		mNumElements++;
	}
	if (vertexFlags & VU_UV1) {
		mElements[mNumElements] = VertexElement(offset, VU_UV1, VT_FLOAT2);
		offset += (2 * sizeof(float));
		mNumElements++;
	}
	mVertexSize = offset;
}

//--------------------------------------------
// RenderBlock
//--------------------------------------------
void gfxCreateRenderBlock(RenderBlock* block, const VertexFormat* fmt, void* vertexBuffer, size_t vertexCount, u32* indexBuffer, size_t indexCount)
{
	gfxCreateRenderBlock(block, fmt, vertexCount, indexCount);

	void* verts = gfxLockVertices(block);
	memcpy(verts, vertexBuffer, vertexCount * fmt->GetVertexSize());
	gfxUnlockVertices(block);

	void* indices = gfxLockIndices(block);
	memcpy(indices, indexBuffer, indexCount * sizeof(u32));
	gfxUnlockIndices(block);
}

void gfxRender(PushBuffer* pb, const RenderBlock* block, u32 firstIndex, u32 numIndices) {
	u32 flags = block->mFormat->GetVertexFlags();
	pb->SetVertexFormat(flags);
	pb->SetVertexBuffer(block->mVB);

	size_t offset = 0;
	u32 index = 0;
	if (flags & VU_POS) {
		pb->SetVertexAttribPointer(index++, reinterpret_cast<char *>( 0 ) + offset);
		offset += (3 * sizeof(float));
	}
	if (flags & VU_NRM) {
		pb->SetVertexAttribPointer(index++, reinterpret_cast<char *>( 0 ) + offset);
		offset += (3 * sizeof(float));
	}
	if (flags & VU_TAN) {
		pb->SetVertexAttribPointer(index++, reinterpret_cast<char *>( 0 ) + offset);
		offset += (4 * sizeof(float));
	}
	if (flags & VU_COL) {
		pb->SetVertexAttribPointer(index++, reinterpret_cast<char *>( 0 ) + offset);
		offset += (4 * sizeof(unsigned char));
	}
	if (flags & VU_UV0) {
		pb->SetVertexAttribPointer(index++, reinterpret_cast<char *>( 0 ) + offset);
		offset += (2 * sizeof(float));
	}
	if (flags & VU_UV1) {
		pb->SetVertexAttribPointer(index++, reinterpret_cast<char *>( 0 ) + offset);
		offset += (2 * sizeof(float));
	}
	pb->SetIndexBuffer(block->mIB);
	pb->DrawIndexedPrim(firstIndex, block->mNumIndices);
	//glDrawElements(GL_TRIANGLES, (GLsizei) mNumIndices, GL_UNSIGNED_SHORT, static_cast<char*>( 0 ));
	//glDrawRangeElementsARB(GL_TRIANGLES, firstIndex, firstIndex + numIndices, numIndices, GL_UNSIGNED_SHORT, static_cast<char*>( 0 ) + sizeof(GL_UNSIGNED_SHORT) * firstIndex);

	pb->DisableVertexAttributes();
}

//--------------------------------------------
// PushBuffer
//--------------------------------------------
extern size_t CommitClear(PushBuffer::Slot* slots);

extern size_t CommitTexture(PushBuffer::Slot* slots);
extern size_t CommitRenderState(PushBuffer::Slot* slots);
extern size_t CommitBlendMode(PushBuffer::Slot* slots);

extern size_t CommitVertexBuffer(PushBuffer::Slot* slots);
extern size_t CommitIndexBuffer(PushBuffer::Slot* slots);

extern size_t CommitVertexFormat(PushBuffer::Slot* slots);
extern size_t CommitVertexAttribPointer(PushBuffer::Slot* slots);
extern size_t CommitDisableVertexAttributes(PushBuffer::Slot* slots);

extern size_t CommitDrawIndexedPrim(PushBuffer::Slot* slots);
extern size_t CommitViewport(PushBuffer::Slot* slots);
extern size_t CommitVertexProgram(PushBuffer::Slot* slots);
extern size_t CommitFragmentProgram(PushBuffer::Slot* slots);
extern size_t CommitVertexProgramConstant(PushBuffer::Slot* slots);

extern size_t CommitEnableRenderTarget(PushBuffer::Slot* slots);
extern size_t CommitDisableRenderTarget(PushBuffer::Slot* slots);
extern size_t CommitBindRenderTarget(PushBuffer::Slot* slots);
extern size_t CommitUnbindRenderTarget(PushBuffer::Slot* slots);

extern size_t CommitVertexAttrib4f(PushBuffer::Slot* slots);
extern size_t CommitDrawBegin(PushBuffer::Slot* slots);
extern size_t CommitDrawEnd(PushBuffer::Slot* slots);

void PushBuffer::Clear(u32 color, u32 bufferFlags) {
	mSlots[mCurrSlot++].func = &CommitClear;
	mSlots[mCurrSlot++].i = color;
	mSlots[mCurrSlot++].i = bufferFlags;
}

void PushBuffer::SetRenderState(RenderState state, bool enabled) {
	mSlots[mCurrSlot++].func = &CommitRenderState;
	mSlots[mCurrSlot++].i = state;
	mSlots[mCurrSlot++].b = enabled;
}

void PushBuffer::SetFrontFace(CullState cullState) {
/*	mCmds[mCurrCmd].opcode = OPCODE_SETFRONTFACE;
	mCmds[mCurrCmd].operand[0].i = cullState;
	mCurrCmd++;*/
}

void PushBuffer::SetBlendMode(BlendMode blendMode) {
	mSlots[mCurrSlot++].func = &CommitBlendMode;
	mSlots[mCurrSlot++].i = blendMode;
}

void PushBuffer::SetViewport(int x, int y, int width, int height) {
	mSlots[mCurrSlot++].func = &CommitViewport;
	mSlots[mCurrSlot++].i = x;
	mSlots[mCurrSlot++].i = y;
	mSlots[mCurrSlot++].i = width;
	mSlots[mCurrSlot++].i = height;
}

void PushBuffer::SetTexture(u32 texUnit, Texture* tex) {
	mSlots[mCurrSlot++].func = &CommitTexture;
	mSlots[mCurrSlot++].i = texUnit;
	mSlots[mCurrSlot++].i = tex->handle;
	mSlots[mCurrSlot++].i = tex->format;
	mSlots[mCurrSlot++].i = tex->control1;
	mSlots[mCurrSlot++].i = tex->filter;
}

void PushBuffer::SetVertexFormat(u32 format) {
	mSlots[mCurrSlot++].func = &CommitVertexFormat;
	mSlots[mCurrSlot++].i = format;
}

void PushBuffer::SetVertexBuffer(u32 handle) {
	mSlots[mCurrSlot++].func = &CommitVertexBuffer;
	mSlots[mCurrSlot++].i = handle;
}

void PushBuffer::SetIndexBuffer(u32 handle) {
	mSlots[mCurrSlot++].func = &CommitIndexBuffer;
	mSlots[mCurrSlot++].i = handle;
}

void PushBuffer::SetVertexAttribPointer(u32 index, void* mem) {
	mSlots[mCurrSlot++].func = &CommitVertexAttribPointer;
	mSlots[mCurrSlot++].i = index;
	mSlots[mCurrSlot++].ptr = mem;
}

void PushBuffer::DrawIndexedPrim(u16 firstIndex, u16 numIndices) {
	mSlots[mCurrSlot++].func = &CommitDrawIndexedPrim;
	mSlots[mCurrSlot++].i = firstIndex;
	mSlots[mCurrSlot++].i = numIndices;
}

void PushBuffer::DisableVertexAttributes() {
	mSlots[mCurrSlot++].func = &CommitDisableVertexAttributes;
}

void PushBuffer::SetVertexProgram(VertexProgram* program) {
	mSlots[mCurrSlot++].func = &CommitVertexProgram;
	mSlots[mCurrSlot++].i = program->handle;
}

void PushBuffer::SetVertexProgramConstant(u32 index, float* constant) {
	assert(constant);
	mSlots[mCurrSlot++].func = &CommitVertexProgramConstant;
	mSlots[mCurrSlot++].i = index;
	mSlots[mCurrSlot++].f = constant[0];
	mSlots[mCurrSlot++].f = constant[1];
	mSlots[mCurrSlot++].f = constant[2];
	mSlots[mCurrSlot++].f = constant[3];
}

void PushBuffer::SetFragmentProgram(FragmentProgram* program) {
	mSlots[mCurrSlot++].func = &CommitFragmentProgram;
	mSlots[mCurrSlot++].i = program->handle;
}

void PushBuffer::EnableRenderTarget(RenderTarget* rt) {
	mSlots[mCurrSlot++].func = &CommitEnableRenderTarget;
	mSlots[mCurrSlot++].ptr = rt;
}
void PushBuffer::DisableRenderTarget(RenderTarget* rt) {
	mSlots[mCurrSlot++].func = &CommitDisableRenderTarget;
	mSlots[mCurrSlot++].ptr = rt;
}
void PushBuffer::BindRenderTarget(u32 texUnit, RenderTarget* rt) {
	mSlots[mCurrSlot++].func = &CommitBindRenderTarget;
	mSlots[mCurrSlot++].i = texUnit;
	mSlots[mCurrSlot++].ptr = rt;
}
void PushBuffer::UnbindRenderTarget(u32 texUnit, RenderTarget* rt) {
	mSlots[mCurrSlot++].func = &CommitUnbindRenderTarget;
	mSlots[mCurrSlot++].i = texUnit;
	mSlots[mCurrSlot++].ptr = rt;
}

void PushBuffer::SetVertexAttrib4f(u32 index, float x, float y, float z, float w) {
	mSlots[mCurrSlot++].func = &CommitVertexAttrib4f;
	mSlots[mCurrSlot++].i = index;
	mSlots[mCurrSlot++].f = x;
	mSlots[mCurrSlot++].f = y;
	mSlots[mCurrSlot++].f = z;
	mSlots[mCurrSlot++].f = w;
}
void PushBuffer::DrawBegin(PrimitiveType primType) {
	mSlots[mCurrSlot++].func = &CommitDrawBegin;
	mSlots[mCurrSlot++].i = primType;
}
void PushBuffer::DrawEnd() {
	mSlots[mCurrSlot++].func = &CommitDrawEnd;
}


