#pragma once

#include "core/Defs.h"

typedef unsigned int VBufferU3D;
typedef unsigned int IBufferU3D;

void gfxInit();
void gfxShutdown();

void gfxSetViewport(int x, int y, int width, int height);

enum VertexFlags {
	VU_POS = 1,	// position	(3 floats)
	VU_NRM = 2,	// normal	(3 floats)
	VU_TAN = 4,	// tangent	(4 floats)
	VU_COL = 8,	// color	(1 uint32)
	VU_UV0 = 16,// texture	(2 floats)
	VU_UV1 = 32	// lightmap	(2 floats)
};

enum VertexElementType {
	VT_FLOAT2,
	VT_FLOAT3,
	VT_FLOAT4,
	VT_COLOR4,
	VT_UBYTE4
};

struct VertexElement {
	size_t offset;
	VertexFlags usageFlag;
	VertexElementType type;

	VertexElement() : offset(0), usageFlag(VU_POS), type(VT_FLOAT3) {}
	VertexElement(size_t offs, VertexFlags usage, VertexElementType t) : offset(offs), usageFlag(usage), type(t) {}
};

class VertexFormat {
public:
	VertexFormat() : mNumElements(0), mVertexSize(0), mVertexFlags(0) {}
	VertexFormat(u32 vertexFlags);
	void Set(u32 vertexFlags);
	inline size_t GetNumElements() { return mNumElements; }
	inline const VertexElement& GetElement(size_t idx) const { return mElements[idx]; }
	inline size_t GetVertexSize() const { return mVertexSize; }
	inline u32 GetVertexFlags() const { return mVertexFlags; }
private:
	VertexElement mElements[8];
	size_t mNumElements;
	size_t mVertexSize;	// in bytes
	u32 mVertexFlags;
};

extern VertexFormat* U3DVertexFormat_PosUV;
extern VertexFormat* U3DVertexFormat_PosNrmUV;
extern VertexFormat* U3DVertexFormat_PosNrmUV2;

enum TextureFilter {
	TEXTUREFILTER_MAG_NEAREST = 0,
	TEXTUREFILTER_MAG_LINEAR,
	TEXTUREFILTER_MIN_NEAREST,
	TEXTUREFILTER_MIN_NEAREST_MIPMAP_NEAREST,
	TEXTUREFILTER_MIN_NEAREST_MIPMAP_LINEAR,
	TEXTUREFILTER_MIN_LINEAR,
	TEXTUREFILTER_MIN_LINEAR_MIPMAP_NEAREST,
	TEXTUREFILTER_MIN_LINEAR_MIPMAP_LINEAR
};

enum TextureAddress {
	TEXTUREADDRESS_WRAP = 1,
	TEXTUREADDRESS_BORDER,
	TEXTUREADDRESS_CLAMPTOEDGE,
	TEXTUREADDRESS_MIRROR
};

#define TEXTURE_FORMAT(DIM) ( ((DIM)<<0) )
#define TEXTURE_CONTROL1(RWRAP, TWRAP, SWRAP) ( ((RWRAP)<<6) | ((TWRAP)<<3) | ((SWRAP)<<0) )
#define TEXTURE_FILTER(MAG, MIN) ( ((MAG)<<8) | ((MIN)<<0) )

struct Texture {
	u32 handle;
	u32 format;
	u32 control1;
	u32 filter;
};

enum BufferFlags {
	BUFFER_COLOR = 1,
	BUFFER_DEPTH = 2,
	BUFFER_STENCIL = 4,
};

enum RenderTargetFlags {
	RENDERTARGET_RGBA = 1,
	RENDERTARGET_DEPTH = 2,
	RENDERTARGET_STENCIL = 4,
	RENDERTARGET_FLOAT = 8,
	RENDERTARGET_MULTISAMPLE_2 = 16,
	RENDERTARGET_MULTISAMPLE_4 = 32,
};

struct RenderTargetData;

struct RenderTarget {
	uint width;
	uint height;
	struct RenderTargetData* data;
};

extern void R_CreateRenderTarget(RenderTarget* rt, uint width, uint height, uint flags);
extern void R_ReleaseRenderTarget(RenderTarget* rt);

enum RenderState {
	RENDERSTATE_DEPTHTEST = 1,
	RENDERSTATE_DEPTHWRITE,
	RENDERSTATE_CULLFACE,
	RENDERSTATE_BLEND,
	//RENDERSTATE_GAMMACORRECTION,
};

enum CullState {
	FRONTFACE_CCW,
	FRONTFACE_CW
};

enum BlendMode
{
	//BLEND_MASKED,	///< Final color = Source color if OpacityMask < OpacityMaskClipValue, otherwise the pixel is discarded.
	BLEND_ALPHA,	///< Final color = Source color * Opacity + Dest color * (1 - Opacity).
	BLEND_MAD,		///< Final color = Source color * Opacity + Dest color.
	BLEND_ADD,		///< Final color = Source color + Dest color.
	BLEND_MUL,		///< Final color = Source color * Dest color.
	BLEND_INVMUL,	///< Final color = (1 - Source color) * Dest color.
	//Modulate2X	///< Final color = 2 * Source color * Dest color.
	BLEND_SUB,		///< Final color = Dest color - Source color.
	//BLEND_LIGHTEN,	///< Final color = Max(Source color, Dest color).
	//BLEND_DARKEN,	///< Final color = Min(Source color, Dest color).
	BLEND_SCREEN,	///< Final color = Source color * InvDestColor + Dest color.
};

enum PrimitiveType {
	PRIM_TRILIST = 1,
	PRIM_QUADLIST = 2,
};

enum ShaderType {
	SHADER_VERTEX = 0,
	SHADER_FRAGMENT,
};

class VertexProgram {
public:
	u32 handle;
};

class FragmentProgram {
public:
	void SetConstant(u32 index, float* constant);
	u32 handle;
};

class PushBuffer {
public:
	PushBuffer() : mCurrSlot(0) {}

	inline void Reset() { mCurrSlot = 0; }

	void Clear(u32 color, u32 bufferFlags);
	void SetRenderState(RenderState, bool enabled);
	void SetFrontFace(CullState cullState);
	void SetBlendMode(BlendMode blendMode);
	void SetViewport(int x, int y, int width, int height);

	void SetTexture(u32 texUnit, Texture* tex);
	//InvalidateTextureCachedState(u32 texUnit);

	void SetVertexFormat(u32 format);
	void SetVertexBuffer(u32 handle);
	void SetIndexBuffer(u32 handle);
	//SetIndexBase(void* mem);
	void SetVertexAttribPointer(u32 index, void* mem);
	void DrawBegin(PrimitiveType primType);
	void DrawIndexedPrim(u16 firstIndex, u16 numIndices);
	void DrawEnd();
	void DisableVertexAttributes();

	void SetVertexAttrib4f(u32 index, float x, float y, float z, float w);

	void SetVertexProgram(VertexProgram* program);
	void SetVertexProgramConstant(u32 index, float* constant);

	void SetFragmentProgram(FragmentProgram* program);
	//void RefreshFragmentProgram(FragmentProgram* program);

	void EnableRenderTarget(RenderTarget* rt);
	void DisableRenderTarget(RenderTarget* rt);
	void BindRenderTarget(u32 texUnit, RenderTarget* rt);
	void UnbindRenderTarget(u32 texUnit, RenderTarget* rt);

	struct Slot {
		union {
			u32 i;
			float f;
			bool b;
			void* ptr;
			size_t (*func)(Slot*);
		};
	};

	Slot mSlots[4096];
	size_t mCurrSlot;
};

void gfxCommitPushBuffer(PushBuffer* pb);

struct RenderBlock {
	VBufferU3D mVB;
	IBufferU3D mIB;
	const VertexFormat* mFormat;
	size_t mNumIndices;

	RenderBlock() : mFormat(0), mVB(-1), mIB(-1), mNumIndices(0) {}
};

bool  gfxCreateRenderBlock(RenderBlock* block, const VertexFormat* fmt, size_t numVerts, size_t numIndices, bool isDynamic = false);
void  gfxCreateRenderBlock(RenderBlock* block, const VertexFormat* fmt, void* vertexBuffer, size_t vertexCount, u32* indexBuffer, size_t indexCount);
void  gfxReleaseRenderBlock(RenderBlock* block);
void* gfxLockVertices(RenderBlock* block);
void  gfxUnlockVertices(RenderBlock* block);
void* gfxLockIndices(RenderBlock* block);
void  gfxUnlockIndices(RenderBlock* block);

void gfxRender(PushBuffer* pb, const RenderBlock* block, u32 firstIndex, u32 numIndices);

u32  gfxUploadTexture2D(void *data, int x, int y, u32 flags);
void gfxReleaseTexture(u32 handle);

u32  gfxCreateShaderProgram(ShaderType type, const char* source);
void gfxReleaseShaderProgram(u32 handle);
