#include "RenderAPI.h"
#include "OpenGL_win32.h"

VertexFormat* U3DVertexFormat_PosUV = NULL;
VertexFormat* U3DVertexFormat_PosNrmUV = NULL;
VertexFormat* U3DVertexFormat_PosNrmUV2 = NULL;

void gfxInit()
{
	gfxInitExtensions();

	gfxCgInit();

	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);

	glEnable(GL_CULL_FACE);

	glDisable(GL_LIGHTING);
	glShadeModel(GL_SMOOTH);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	U3DVertexFormat_PosUV = new VertexFormat(VU_POS|VU_UV0);
	U3DVertexFormat_PosNrmUV = new VertexFormat(VU_POS|VU_NRM|VU_UV0);
	U3DVertexFormat_PosNrmUV2 = new VertexFormat(VU_POS|VU_NRM|VU_UV0|VU_UV1);
}

void gfxShutdown()
{
	if (U3DVertexFormat_PosUV) delete U3DVertexFormat_PosUV;
	if (U3DVertexFormat_PosNrmUV)delete U3DVertexFormat_PosNrmUV;
	if (U3DVertexFormat_PosNrmUV2) delete U3DVertexFormat_PosNrmUV2;
	U3DVertexFormat_PosUV = NULL;
	U3DVertexFormat_PosNrmUV = NULL;
	U3DVertexFormat_PosNrmUV2 = NULL;

	gfxCgShutdown();

	gfxDeleteContext();
}

void gfxSetViewport(int x, int y, int width, int height) {
	glViewport(x, y, width, height);
}

//--------------------------------------------
// RenderBlock
//--------------------------------------------
bool gfxCreateRenderBlock(RenderBlock* block, const VertexFormat* fmt, size_t numVerts, size_t numIndices, bool isDynamic)
{
	block->mFormat = fmt;
	block->mNumIndices = numIndices;
	if ((fmt->GetVertexSize() > 0) && (numVerts > 0)) {
		glGenBuffersARB(1, &block->mVB);
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, block->mVB);
		size_t size = fmt->GetVertexSize() * numVerts;

		if (!isDynamic) glBufferDataARB(GL_ARRAY_BUFFER_ARB, size, NULL, GL_STATIC_DRAW_ARB);
		else glBufferDataARB(GL_ARRAY_BUFFER_ARB, size, NULL, GL_DYNAMIC_DRAW_ARB);
	}
	if (numIndices > 0) {
		glGenBuffersARB(1, &block->mIB);
		glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, block->mIB);
		//size_t size = sizeof(u16) * numIndices;
		size_t size2 = (sizeof(u32) * numIndices);
		if (!isDynamic) glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, size2, NULL, GL_STATIC_DRAW_ARB);
		else glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, size2, NULL, GL_DYNAMIC_DRAW_ARB);
	}
	return true;
}

void gfxReleaseRenderBlock(RenderBlock *block) {
	if (block->mVB > 0) glDeleteBuffersARB(1, &block->mVB);
	if (block->mIB > 0) glDeleteBuffersARB(1, &block->mIB);
}

void* gfxLockVertices(RenderBlock *block) {
	void *ptr = NULL;
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, block->mVB);
	ptr = glMapBufferARB(GL_ARRAY_BUFFER_ARB, GL_WRITE_ONLY);
	return ptr;
}

void gfxUnlockVertices(RenderBlock *block) {
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, block->mVB);
	glUnmapBufferARB(GL_ARRAY_BUFFER_ARB);
}

void* gfxLockIndices(RenderBlock *block) {
	void *ptr = NULL;
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, block->mIB);
	ptr = glMapBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, GL_WRITE_ONLY);
	return ptr;
}

void gfxUnlockIndices(RenderBlock *block) {
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, block->mIB);
	glUnmapBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB);
}


//--------------------------------------------
// PushBuffer
//--------------------------------------------

// state
static VertexFormat vertexFormat = NULL;

size_t CommitClear(PushBuffer::Slot* slots) {
	u32 color = (u32)slots[0].i;
	u32 bufferFlags = (u32)slots[1].i;
	
	float a = ((color & 0xff000000) >> 24) / 255.0f;
	float r = ((color & 0x00ff0000) >> 16) / 255.0f;
	float g = ((color & 0x0000ff00) >>  8) / 255.0f;
	float b = (color & 0xff) / 255.0f;

	u32 glflags = 0;
	if (bufferFlags & BUFFER_COLOR) glflags |= GL_COLOR_BUFFER_BIT;
	if (bufferFlags & BUFFER_DEPTH) glflags |= GL_DEPTH_BUFFER_BIT;
	if (bufferFlags & BUFFER_STENCIL) glflags |= GL_STENCIL_BUFFER_BIT;

	glClearColor(r, g, b, a);
	glClear(glflags);
	return 2;
}

//void CommitTexture(u32 texUnit, u32 handle, u32 format, u32 control1, u32 filter) {
size_t CommitTexture(PushBuffer::Slot* slots) {
	u32 texUnit = (u32)slots[0].i;
	u32 handle = (u32)slots[1].i;
	u32 format = (u32)slots[2].i;
	u32 control1 = (u32)slots[3].i;
	u32 filter = (u32)slots[4].i;

	GLenum target = GL_TEXTURE_2D;
	if (format == 3) target = GL_TEXTURE_3D;
	glActiveTextureARB(GL_TEXTURE0_ARB+texUnit);
	glClientActiveTextureARB(GL_TEXTURE0+texUnit);
	glEnable(target);
	//glActiveTexture()
	glBindTexture(target, handle);

	// TODO: control1
	glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_REPEAT);

	u32 minFilter = GL_NEAREST;
	u32 magFilter = GL_NEAREST;
	if (((filter >> 8) & 0xff) == TEXTUREFILTER_MAG_LINEAR) magFilter = GL_LINEAR;

	if ((filter & 0xff) == TEXTUREFILTER_MIN_LINEAR) minFilter = GL_LINEAR;
	else if ((filter & 0xff) == TEXTUREFILTER_MIN_LINEAR_MIPMAP_LINEAR) minFilter = GL_LINEAR_MIPMAP_LINEAR;

	glTexParameteri(target, GL_TEXTURE_MAG_FILTER, magFilter);
	glTexParameteri(target, GL_TEXTURE_MIN_FILTER, minFilter);

	return 5;
}

size_t CommitRenderState(PushBuffer::Slot* slots) {
	RenderState state = (RenderState)slots[0].i;
	bool enabled = slots[1].b;

	if (state == RENDERSTATE_DEPTHTEST) {
		if (enabled) glEnable(GL_DEPTH_TEST);
		else glDisable(GL_DEPTH_TEST);
	}
	else if (state == RENDERSTATE_BLEND) {
		if (enabled) glEnable(GL_BLEND);
		else glDisable(GL_BLEND);
	}
	else if (state == RENDERSTATE_CULLFACE) {
		if (enabled) glEnable(GL_CULL_FACE);
		else glDisable(GL_CULL_FACE);
	}
	else if (state == RENDERSTATE_DEPTHWRITE) glDepthMask(enabled);

	return 2;	// number of consumed slots
}

//void CommitBlendMode(BlendMode blendMode) {
size_t CommitBlendMode(PushBuffer::Slot* slots) {
	BlendMode blendMode = (BlendMode)slots[0].i;

	if (blendMode == BLEND_ADD) glBlendFunc(GL_ONE, GL_ONE);
	else if (blendMode == BLEND_MAD) glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	else if (blendMode == BLEND_ALPHA) glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	else if (blendMode == BLEND_MUL) glBlendFunc(GL_DST_COLOR, GL_ZERO);
	else if (blendMode == BLEND_INVMUL) glBlendFunc(GL_ZERO, GL_ONE_MINUS_SRC_COLOR);
	else if (blendMode == BLEND_SCREEN) glBlendFunc(GL_ONE_MINUS_DST_COLOR, GL_ONE);
//	else if (blendMode == BLEND_LIGHTEN) { glBlendFunc(GL_ONE, GL_ONE); }
//	else if (blendMode == BLEND_DARKEN) { glBlendFunc(GL_ONE, GL_ONE); }

	return 1;
}

size_t CommitVertexBuffer(PushBuffer::Slot* slots) {
	u32 handle = (u32)slots[0].i;
	assert(handle > 0);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, handle);
	return 1;
}

size_t CommitIndexBuffer(PushBuffer::Slot* slots) {
	u32 handle = (u32)slots[0].i;
	assert(handle > 0);
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, handle);
	return 1;
}

size_t CommitVertexFormat(PushBuffer::Slot* slots) {
	u32 flags = (u32)slots[0].i;
	if (flags != vertexFormat.GetVertexFlags()) vertexFormat.Set(flags);
	return 1;
}

//void CommitVertexAttribPointer(VertexFormat& vertexFormat, u32 index, void* mem) {
size_t CommitVertexAttribPointer(PushBuffer::Slot* slots) {
	u32 index = (u32)slots[0].i;
	void* mem = (void*)slots[1].ptr;

	assert(index < vertexFormat.GetNumElements());

	GLsizei stride = (GLsizei) vertexFormat.GetVertexSize();

	const VertexElement& vtxElem = vertexFormat.GetElement(index);
	u32 flags = vtxElem.usageFlag;

	//glEnableVertexAttribArray(index);

	if (flags & VU_POS) {
//		glEnableClientState(GL_VERTEX_ARRAY);
//		glVertexPointer(3, GL_FLOAT, stride, mem);
		glEnableVertexAttribArray(index);
		glVertexAttribPointer(index, 3, GL_FLOAT, false, stride, mem);
	}
	else if (flags & VU_NRM) {
//		glEnableClientState(GL_NORMAL_ARRAY);
//		glNormalPointer(GL_FLOAT, stride, mem);
		glEnableVertexAttribArray(index);
		glVertexAttribPointer(index, 3, GL_FLOAT, false, stride, mem);
	}
	else if (flags & VU_TAN) {
//		glActiveTextureARB(GL_TEXTURE2_ARB);
		//glClientActiveTextureARB(GL_TEXTURE2);
		//glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		//glTexCoordPointer(4, GL_FLOAT, stride, mem);
//		glEnableVertexAttribArray(index);
//		glVertexAttribPointer(index, 4, GL_FLOAT, false, stride, mem);
	}
	else if (flags & VU_COL) {
//		glEnableClientState(GL_COLOR_ARRAY);
//		glColorPointer(4, GL_UNSIGNED_BYTE, stride, mem);
//		glEnableVertexAttribArray(index);
//		glVertexAttribPointer(index, 4, GL_UNSIGNED_BYTE, false, stride, mem);
	}
	else if (flags & VU_UV0) {
		glActiveTextureARB(GL_TEXTURE0_ARB);
		glClientActiveTextureARB(GL_TEXTURE0);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(2, GL_FLOAT, stride, mem);
//		glEnableVertexAttribArray(index);
//		glVertexAttribPointer(index, 2, GL_FLOAT, 0, stride, mem);
	}
	else if (flags & VU_UV1) {
		glActiveTextureARB(GL_TEXTURE1_ARB);
		glClientActiveTextureARB(GL_TEXTURE1);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(2, GL_FLOAT, stride, mem);
//		glEnableVertexAttribArray(index);
//		glVertexAttribPointer(index, 2, GL_FLOAT, 0, stride, mem);
	}

	return 2;
}

//void CommitDisableVertexAttributes(u32 flags) {
size_t CommitDisableVertexAttributes(PushBuffer::Slot* slots) {
	u32 flags = vertexFormat.GetVertexFlags();

	/*
	if (flags & VU_NRM) glDisableClientState(GL_NORMAL_ARRAY);
	if (flags & VU_POS) glDisableClientState(GL_VERTEX_ARRAY);
	if (flags & VU_COL) glDisableClientState(GL_COLOR_ARRAY);
	if (flags & VU_TAN) {
		glActiveTextureARB(GL_TEXTURE2_ARB);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	}
	if (flags & VU_UV1) {
		glActiveTextureARB(GL_TEXTURE1_ARB);
		glClientActiveTextureARB(GL_TEXTURE1);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	}
	if (flags & VU_UV0) {
		glActiveTextureARB(GL_TEXTURE0_ARB);
		glClientActiveTextureARB(GL_TEXTURE1);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	}*/
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,0);
	int numArrays = CountBitsInMask(flags);
	for (int i=0; i<numArrays; i++)glDisableVertexAttribArray(numArrays-1-i);

//**************************************************
//	glDisable(GL_FRAGMENT_PROGRAM_ARB); 
//	glDisable(GL_VERTEX_PROGRAM_ARB);
//**************************************************

	return 0;
}

size_t CommitDrawIndexedPrim(PushBuffer::Slot* slots) {
	u32 firstIndex = (u32)slots[0].i;
	u32 numIndices = (u32)slots[1].i;
	//glDrawElements(GL_TRIANGLES, (GLsizei) numIndices, GL_UNSIGNED_SHORT, static_cast<char*>( 0 ));
	glDrawElements(GL_TRIANGLES, (GLsizei) numIndices, GL_UNSIGNED_INT, static_cast<void*>( 0 ));
	//glDrawRangeElementsARB(GL_TRIANGLES, firstIndex, firstIndex + numIndices, numIndices, GL_UNSIGNED_SHORT, static_cast<char*>( 0 ) + sizeof(GL_UNSIGNED_SHORT) * firstIndex);
	return 2;
}

size_t CommitViewport(PushBuffer::Slot* slots) {
	glViewport(slots[0].i, slots[1].i, slots[2].i, slots[3].i);
	return 4;
}

size_t CommitVertexProgram(PushBuffer::Slot* slots) {
	glEnable(GL_VERTEX_PROGRAM_ARB);
	glBindProgramARB(GL_VERTEX_PROGRAM_ARB, slots[0].i);
	return 1;
}

size_t CommitFragmentProgram(PushBuffer::Slot* slots) {
	glEnable(GL_FRAGMENT_PROGRAM_ARB);
	glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, slots[0].i);
	return 1;
}

size_t CommitVertexProgramConstant(PushBuffer::Slot* slots) {
	int index = slots[0].i;
	float params[4];
	params[0] = slots[1].f;
	params[1] = slots[2].f;
	params[2] = slots[3].f;
	params[3] = slots[4].f;
	glProgramLocalParameter4fvARB(GL_VERTEX_PROGRAM_ARB, index, params);
	//glProgramEnvParameter4fvARB(GL_VERTEX_PROGRAM_ARB, index, params);
	return 5;
}

size_t CommitEnableRenderTarget(PushBuffer::Slot* slots) {
	gfxEnableRenderTarget((RenderTarget*)slots[0].ptr);
	return 1;
}

size_t CommitDisableRenderTarget(PushBuffer::Slot* slots) {
	gfxDisableRenderTarget((RenderTarget*)slots[0].ptr);
	return 1;
}

size_t CommitBindRenderTarget(PushBuffer::Slot* slots) {
	gfxBindRenderTarget(slots[0].i, (RenderTarget*)slots[1].ptr);
	return 2;
}

size_t CommitUnbindRenderTarget(PushBuffer::Slot* slots) {
	gfxUnbindRenderTarget(slots[0].i, (RenderTarget*)slots[1].ptr);
	return 2;
}

size_t CommitVertexAttrib4f(PushBuffer::Slot* slots) {
	glVertexAttrib4f(slots[0].i, slots[1].f, slots[2].f, slots[3].f, slots[4].f);
	return 5;
}
size_t CommitDrawBegin(PushBuffer::Slot* slots) {
	GLenum mode = GL_TRIANGLES;
	if (slots[0].i == PRIM_QUADLIST) mode = GL_QUADS;
	glBegin(mode);
	return 1;
}
size_t CommitDrawEnd(PushBuffer::Slot* slots) {
	glEnd();
	return 0;
}

void gfxCommitPushBuffer(PushBuffer* pb) {
	size_t s = 0;
	while (s < pb->mCurrSlot) {
		s += pb->mSlots[s].func(&pb->mSlots[s+1]);
		s++;
	}
}

//--------------------------------------------
// Texture management
//--------------------------------------------
void gfxReleaseTexture(u32 handle) {
	glDeleteTextures(1, (unsigned int*)&handle);
}

u32 gfxUploadTexture2D(void *data, int x, int y, u32 flags) {
	GLuint tex;

	if (!data) return 0;

	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);

	/*
	if (flags & 1)
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
	else
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);

	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	*/

	int level=0;

	do
	{
		glTexImage2D(GL_TEXTURE_2D, level, GL_RGBA, x, y, 0, GL_BGRA, GL_UNSIGNED_BYTE, data);

		int nx= x>>1;
		int ny= y>>1;

		unsigned int *dst= (unsigned int *)data;
		unsigned int *src1= (unsigned int *)data;
		for (int i=0;i<ny;i++)
		{
			unsigned int *src2= src1 + x;
			for (int j=0;j<nx;j++)
			{
				unsigned int c;
				unsigned int c1= *src1++;
				unsigned int c2= *src1++;
				unsigned int c3= *src2++;
				unsigned int c4= *src2++;

				int a= ((c1>>24&0xff) + (c2>>24&0xff) + (c3>>24&0xff) + (c4>>24&0xff)) >> 2;
				int r= ((c1>>16&0xff) + (c2>>16&0xff) + (c3>>16&0xff) + (c4>>16&0xff)) >> 2;
				int g= ((c1>> 8&0xff) + (c2>> 8&0xff) + (c3>> 8&0xff) + (c4>> 8&0xff)) >> 2;
				int b= ((c1    &0xff) + (c2    &0xff) + (c3    &0xff) + (c4    &0xff)) >> 2;
				c= (a<<24)+(r<<16)+(g<<8)+b;

				*dst++= c;
			}
			src1+=(x<<1)-(nx<<1);
		}

		x= nx;
		y= ny;

		level++;
	} while (x>0 && y>0 && (flags & 1));

	return (u32)tex;
}

u32 gfxCreateShaderProgram(ShaderType shaderType, const char* source) {
	GLuint ID;
	GLenum type = GL_VERTEX_PROGRAM_ARB;
	if (shaderType == SHADER_FRAGMENT) type = GL_FRAGMENT_PROGRAM_ARB;

	size_t length = strlen(source);

	glGenProgramsARB(1, &ID);
	glBindProgramARB(type, ID);
	glProgramStringARB(type, GL_PROGRAM_FORMAT_ASCII_ARB, (GLsizei)length, source);

	if(GL_INVALID_OPERATION == glGetError()) {
		//GLubyte const* errorString = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
		//std::cout << "error while loading shader" << std::endl << errorString << std::endl;
		return 0;
	} 
	return ID;
}

void gfxReleaseShaderProgram(u32 handle) {
	glDeleteProgramsARB(1, &handle);
}
