#include "FileIO.h"
#include <io.h>

FileIO::FileList FileIO::GetFileList(const std::string& p, bool recursive, const std::string& wildcard)
{
	std::string path = p; 

	// add '/' to end if there isn't yet
	if( !path.empty() && path[path.size()-1]!='\\' && path[path.size()-1]!='/' )
		path += "/";

	FileList files;

	// non-listing directory?
	//if( isNonListing( path ) ) return files;

	// find first
	_finddata_t file;
	int handle = _findfirst( (path+wildcard).c_str(), &file );
	int ret = handle;

	// while found
	while( ret >= 0 ) {
		if( !( file.attrib & _A_SUBDIR ) ) {
			files.push_back( path + file.name );
		} else if( recursive ) {
			std::string subDir( file.name );

			StringUtils::toLower( subDir );
			
			if( subDir!="." && subDir!=".." && subDir!="svn" )
			{
				FileList l = GetFileList(path + subDir, true);
				files.splice( files.end(), l );
			}
		}
		ret = _findnext( handle, &file );
	}

	return files;
}

void* FileIO::LoadFile(const std::string& filename, unsigned int& size)
{
	FILE* fp;
	if(!(fp = fopen(filename.c_str(), "rb"))) return NULL;

	fseek(fp, 0, SEEK_END);
	size = ftell(fp);
	rewind(fp);

	unsigned char* buffer = new unsigned char[size];
	fread(buffer, 1, size, fp);
	fclose(fp);

	return buffer;
}

std::string FileIO::GetFileExtension(const std::string& filename)
{
	std::string::size_type pos = filename.find_last_of('.');
	if(std::string::npos != pos)
	{
		std::string ext = filename.substr(pos + 1);
		StringUtils::toLower(ext);
		return ext;
	}
	return "";
}

std::string FileIO::ConvertFilename(const std::string& _filename)
{
	char *fname;

	int len = (int)_filename.length();

	fname = new char[len+1];
	strcpy(fname, _filename.c_str());

    _strlwr(fname);
    for(int i=0; i < len; i++) 
	{
		if(fname[i] == '\\') fname[i] = '/';
	}
	std::string filename = (std::string) fname;
	delete fname;

	return filename; //(std::string) fname;
}

std::string FileIO::GetFilenameNoPath(const std::string& fname)
{
	std::string filename = ConvertFilename(fname);

    // TODO: Do we need to handle e.g. D:?
    std::string::size_type pos = filename.find_last_of('/');
    if(std::string::npos != pos) {
        return filename.substr(pos + 1);
    } else {
        return filename;
    }
}

//----------------------------------------------------------------------------
// Stream
//----------------------------------------------------------------------------

Stream::Stream():
isUserBuffer(false)
{
	// test endian coding
	u32 test = 0x01020304;
	unsigned char *data = (unsigned char*)&test;
	if(data[0]==4 && data[1]==3 && data[2]==2 && data[3]==1) m_cpuEndian = EM_LITTLE ;
	if(data[0]==1 && data[1]==2 && data[2]==3 && data[3]==4) m_cpuEndian = EM_BIG;
	
	// set the streams mode to cpu mode
	m_streamEndian = m_cpuEndian;
	
	close();
}

Stream::~Stream()
{
	isUserBuffer = false;
	close();
}

bool Stream::open(void *buffer, size_t bufferSize, EndianMode endianMode)
{
	close();
	
	if(buffer)
	{
		m_bufferStart = (unsigned char*) buffer;
		isUserBuffer = true;
		m_size = bufferSize;
	}
	else
	{
		if(!realloc()) return false;
		isUserBuffer = false;
	}
		
	m_buffer = m_bufferStart;		
	m_streamEndian = endianMode;
	return true;
}

void Stream::close()
{
	if(isUserBuffer && m_bufferStart)
	{
		delete[] m_bufferStart;
	}
	m_buffer = NULL;
	m_bufferStart = NULL;
	isUserBuffer = false;
	m_size = 0;
}

bool Stream::realloc(size_t additionalAmount)
{
	if(isUserBuffer) return false;
	
	// allocate bigger buffer
	unsigned char *newBuffer = new unsigned char[m_size + additionalAmount];
	if(!newBuffer)
	{
		close();
		return false;
	}
	size_t offset = getPos();		// save current buffer offset
	
	if(m_bufferStart)
	{
		// copy old buffer contents to new buffer
		memcpy(newBuffer, m_bufferStart, offset);
	
		delete[] m_bufferStart;				// free old buffer
	}
	m_bufferStart = newBuffer;			// set new buffer
	m_buffer = m_bufferStart + offset;	// set new pointer to offset
	m_size += additionalAmount;			// update size
	
	return true;
}

void Stream::reserve(size_t s)
{
	if(m_size < s)
	{
		realloc(s - m_size);
	}
}
	
void Stream::advance(int s)
{
	assert((m_buffer + s) <= (m_bufferStart + m_size));  
	assert((m_buffer + s) >= m_bufferStart);  
	m_buffer += s;
}

void Stream::seek(int offset, ESeekMode seekOrigin)
{
    if(seekOrigin == Seek_SET) {
            m_buffer = offset + m_bufferStart;
    }
    else if(seekOrigin == Seek_CUR) {
            m_buffer += offset;
    }
    else if(seekOrigin == Seek_END) {
            m_buffer = offset + m_size + m_bufferStart;
    }
	assert(m_buffer <= (m_bufferStart + m_size));  
	assert(m_buffer >= m_bufferStart);  
}

void Stream::readRaw(u8* buffer, size_t len)
{
	memcpy(buffer, m_buffer, len);
	m_buffer += len;
}

u8 Stream::readU8()
{
	unsigned char data = *m_buffer;
	m_buffer++;
	return data;
}

u16 Stream::readU16()
{
	u16 data = *(u16*)m_buffer;
	m_buffer += sizeof(u16);
	if(m_cpuEndian != m_streamEndian) swap16(&data);
	return data;
}

u32 Stream::readU32()
{
	u32 data = *(u32*)m_buffer;
	m_buffer += sizeof(u32);
	if(m_cpuEndian != m_streamEndian) swap32(&data);
	return data;
}

float Stream::readFloat()
{
	float data = *(float*)m_buffer;
	m_buffer += sizeof(float);
	if(m_cpuEndian != m_streamEndian) swap32(&data);
	return data;
}

void Stream::readString(std::string &dest)
{
	dest.clear();
	
	u32 length = readU32();
	dest.reserve(length);
	dest.assign(reinterpret_cast<std::string::const_pointer>(m_buffer),length);
	m_buffer+=length*sizeof(std::string::value_type);
}

void Stream::readArrayU16(u16* buffer, size_t count)
{
	for(size_t i=0; i < count; i++) {
		buffer[i] = readU16();
	}
}

void Stream::readArrayU32(u32* buffer, size_t count)
{
	for(size_t i=0; i < count; i++) {
		buffer[i] = readU32();
	}
}

void Stream::readArrayFloat(float* buffer, size_t count)
{
	for(size_t i=0; i < count; i++) {
		buffer[i] = readFloat();
	}
}

bool Stream::writeU8(u8 data)
{
	if(getPos()+1 > m_size)
	{
		if(!realloc()) return false;
	}
	
	// update buffer
	*m_buffer++ = data;
	
	return true;
}

bool Stream::writeU16(u16 data)
{
	if(getPos()+sizeof(u16) > m_size)
	{
		if(!realloc()) return false;
	}
	if(m_cpuEndian != m_streamEndian) swap16(&data);
	
	// update buffer
	*(u16*)m_buffer = data;
	m_buffer += sizeof(u16);
	
	return true;
}

bool Stream::writeU32(u32 data)
{
	if(getPos()+sizeof(u32) > m_size)
	{
		if(!realloc()) return false;
	}
	if(m_cpuEndian != m_streamEndian) swap32(&data);
	
	// update buffer
	*(u32*)m_buffer = data;
	m_buffer += sizeof(u32);
	
	return true;
}

bool Stream::writeFloat(float data)
{
	if(getPos()+sizeof(float) > m_size)
	{
		if(!realloc()) return false;
	}
	if(m_cpuEndian != m_streamEndian) swap32(&data);
	
	// update buffer
	*(float*)m_buffer = data;
	m_buffer += sizeof(float);
	
	return true;
}

bool Stream::writeString(const std::string& data)
{
	// string length plus characters
	size_t	strCount = data.size();
	assert(strCount < (size_t)(~(u32)0));
	
    size_t	strSize = strCount * sizeof(std::string::value_type);
    
	size_t	len = strSize + sizeof(u32);
	while(getPos()+len > m_size)
	{
		if(!realloc()) return false;
	}
		
	// update buffer
	writeU32((u32)strCount);
	
	// copy content
	memcpy(m_buffer,data.data(),strSize);
	m_buffer += strSize;
	
	return true;
}

bool Stream::writeRaw(u8* buffer, size_t len)
{
	while(getPos()+len > m_size)
	{
		if(!realloc()) return false;
	}
			
	// copy content
	memcpy(m_buffer, buffer, len);
	m_buffer += len;
	
	return true;
}

void Stream::swap16(void *data)
{
   unsigned char *src = (unsigned char*)data;
   unsigned char t;
   t = src[0]; src[0] = src[1]; src[1] = t;
}

void Stream::swap32(void *data)
{
	u32 v = *(u32*)data;
	*(u32*)data =	((v & 0xFF) << 24 ) |
					((v & 0xFF00) << 8 ) |
					((v & 0xFF0000) >> 8 ) |
					((v & 0xFF000000) >> 24 );
}


//----------------------------------------------------------------------------
// ChunkFile
//----------------------------------------------------------------------------

ChunkFile::ChunkFile() {
	inchunk = 0;
	f = NULL;
	writing = 0;
	chunksize = chunkstart = 0;	
}

bool ChunkFile::Open(File *file, bool forwrite) {
	//if(f) f->Close();
	writing = forwrite;
	f = file;
	return f != NULL;
}

void ChunkFile::Close() {
	//if(f) f->Close();
	f = NULL;
}

int ChunkFile::StartChunk(u32 id) {
	if(writing && f)
	{	
		if(inchunk) EndChunk();
		inchunk = 1;
		chunksize = 0;
		chunkstart = (u32) f->GetPos();		
		f->WriteU32(id);
		f->WriteU32(chunksize);
	}
	return 0;
}

void ChunkFile::EndChunk()
{
	if(f && inchunk)
	{
		inchunk = 0;
		if(writing)
		{
			u32 cur = (u32) f->GetPos();
			chunksize = cur - chunkstart;
			f->Seek(chunkstart + sizeof(u32), Seek_SET);
			f->WriteU32(chunksize);
			f->Seek(cur, Seek_SET);
		}
		else
		{
			f->Seek(chunkstart + chunksize, Seek_SET);
		}
	}
}

u32 ChunkFile::ReadNextChunk(u32 &id)
{
	if(f->Eof()) return 0;
	if(!writing && f)
	{
		if(inchunk) EndChunk();
		chunkstart = (u32) f->GetPos();
		id = f->ReadU32();
		if(f->Eof()) return 0;
		chunksize = f->ReadU32();
		inchunk = 1;
	}
	return chunksize;
}

u32 ChunkFile::ExamineNextChunk(u32 &id) {
	u32 csize = 0;
	if(f->Eof()) return 0;
	if(!writing && f)
	{
		if(inchunk) EndChunk();
		id = f->ReadU32();
		csize = f->ReadU32();
		// undo read (id & size, see above)
		f->Seek((int)f->GetPos() - 2 * sizeof(u32), Seek_SET);
	}
	return csize;
}

