#pragma once
#include "Defs.h"
#include <list>
#include <stdio.h>

//----------------------------------------------------------------------------
// Filesystem functions
//----------------------------------------------------------------------------
class FileIO
{
public:
	typedef std::list<std::string> FileList;

	static FileList GetFileList(const std::string& path, bool recursive, const std::string& wildcard = "*");
	
	static void* LoadFile(const std::string& filename, unsigned int& size);

	static std::string GetFileExtension(const std::string& filename);
	static std::string GetFilenameNoPath(const std::string& filename);
	static std::string ConvertFilename(const std::string& filename);
};

/// Specifies possible seek origins
enum ESeekMode
{
	Seek_SET = 0,
	Seek_CUR,
	Seek_END
};

//----------------------------------------------------------------------------
// Data buffer stream class.
// - supports little and big endian conversion.
//----------------------------------------------------------------------------
class Stream
{
public:
	enum { BLOCK_SIZE = 4096 };

	enum EndianMode
	{
		EM_LITTLE = 0,
		EM_BIG
	};
	
	Stream();
	~Stream();
	
	/// Initialize stream data buffer.
	/// If user-supplied buffer is NULL, an internal resizable buffer is used.	
	bool open(void *buffer = NULL, size_t bufferSize = 0, EndianMode endianMode = EM_LITTLE);
	
	/// Reset stream.
	/// NOTE: This does _not_ free any user-supplied buffers.
	void close();
	
	void	readRaw(u8* buffer, size_t len);
	u8		readU8();
	u16 	readU16();
	u32 	readU32();
	float 	readFloat();
	/// Read a string from the stream. Format: Length (UINT32), Characters.
	void 	readString(std::string &dest);
	void	readArrayU16(u16* buffer, size_t count);
	void	readArrayU32(u32* buffer, size_t count);
	void	readArrayFloat(float* buffer, size_t count);
	
	bool writeRaw(u8* buffer, size_t len);
	bool writeU8(u8 data);
	bool writeU16(u16 data);
	bool writeU32(u32 data);
	bool writeFloat(float data);
	/// Write a string to the stream. Format: Length (UINT32), Characters.
	bool writeString(const std::string &data);
	
	/// returns the current relative writing position within the stream
	size_t getPos() const	{	return (m_buffer - m_bufferStart);		}

	/// Move buffer offset pointer to buffer start
	void rewind()			{	m_buffer = m_bufferStart;	}
	
	/// ensures a space of at least the given size is available
	void reserve(size_t s);
	
	/// advances the write pointer with the given number of bytes
	void advance(int s);
	
	void seek(int offset, ESeekMode seekOrigin);

	/// returns true if reader is at the end of stream
	bool	eof() const		{	return (m_bufferStart + m_size) <=  m_buffer;	}
	
	/// returns a pointer to the beginning of the internal stream data buffer
	unsigned char *getBuffer() const { return m_bufferStart; }
	
protected:
	EndianMode		m_cpuEndian;	///< endianess of host CPU
	EndianMode		m_streamEndian;	///< endianess of stream buffer
	unsigned char	*m_buffer;		///< stream data buffer
	unsigned char	*m_bufferStart;	///< stream data buffer
	
	size_t			m_size;			///< allocated size of stream buffer
	bool			isUserBuffer;	///< Flag: User-provided data buffer

	/// grow data buffer if neccessary
	bool realloc(size_t additionalAmount = BLOCK_SIZE);
	
	/// handle byte-order for 16-bit int
	void swap16(void *data);
	/// handle byte-order for 32-bit int
	void swap32(void *data);
};

//----------------------------------------------------------------------------
// File
//----------------------------------------------------------------------------
class File
{
public:
	enum EFileMode {
		Mode_Read,
		Mode_Write,
		Mode_Append
	};

	virtual ~File() {}

	virtual void Close() = 0;

	EFileMode GetMode() const { return mFileMode; }

	virtual void Read(void * buf, size_t bytes) = 0;
	virtual void Write(const void * buf, size_t bytes) = 0;

	virtual size_t GetPos() = 0;
	virtual void Seek(int offset, ESeekMode seekOrigin) = 0;
	virtual bool Eof() = 0;

	inline u8  ReadU8()  { u8 data;  Read(&data, sizeof(u8));  return data; }
	inline u16 ReadU16() { u16 data; Read(&data, sizeof(u16)); return data; }
	inline u32 ReadU32() { u32 data; Read(&data, sizeof(u32)); return data; }
	inline float ReadFloat() { float data; Read(&data, sizeof(float)); return data; }

	inline void WriteU8(u8 data)   { Write(&data, sizeof(u8));  }
	inline void WriteU16(u16 data) { Write(&data, sizeof(u16)); }
	inline void WriteU32(u32 data) { Write(&data, sizeof(u32)); }
	inline void WriteFloat(float data) { Write((u8*)&data, sizeof(float)); }
protected:
	EFileMode mFileMode;
};

class StdFile : public File {
public:
	StdFile(const char* filename, EFileMode mode) {
		if (mode == Mode_Read) fp = fopen(filename, "rb");
		if (mode == Mode_Write) fp = fopen(filename, "wb");
		if (!fp) return;
		mFileMode = mode;
	}
	~StdFile() { Close(); }

	size_t GetPos() { return ftell(fp); }
	void Seek(int offset, ESeekMode seekOrigin) { fseek(fp, offset, (int)seekOrigin); }
	bool Eof() { return (feof(fp) != 0); }

	// TODO: assert
	void Read(void * buf, size_t bytes) { fread(buf, 1, bytes, fp); }
	void Write(const void * buf, size_t bytes) { fwrite(buf, 1, bytes, fp); }
	void Close() { if(fp) fclose(fp); }
protected:
	FILE* fp;
};

class MemFile : public File {
public:
	MemFile(void *buffer = NULL, size_t bufferSize = 0) {
		s.open(buffer, bufferSize);
	}
	~MemFile() { Close(); }
	
	size_t GetPos() { return s.getPos(); }
	void Seek(int offset, ESeekMode seekOrigin) { s.seek(offset, seekOrigin); }
	bool Eof() { return s.eof(); }

	void Read(void * buf, size_t bytes) { s.readRaw((u8*)buf, bytes); }
	void Write(const void * buf, size_t bytes) { s.writeRaw((u8*)buf, bytes); }
	void Close() { s.close(); }
protected:
	Stream s;
};

//----------------------------------------------------------------------------
// Chunk-based file IO 
// - handles reading/writing of data chunks
//----------------------------------------------------------------------------
struct ChunkFile
{
	File *f;
	bool writing;
	u32 chunksize;
	u32 chunkstart;
	u32 inchunk;

	ChunkFile();

	/// Set stream to use
	/// \param stream File/Memory stream to use (already opened)
	/// \param forwrite True, if used for writing
	bool Open(File* file, bool forwrite = false);
	void Close();

	/// Write chunk header
	/// \param id Chunk identifier (magic)
	int StartChunk(u32 id);

	/// Ends current chunk. Can be used to skip chunk if in read mode.
	void EndChunk();

	/// Jumps to next chunk and gets its header and size
	/// \param id Identifier of next chunk
	/// \return Size of next chunk, 0 if end of file (EOF)
	u32 ReadNextChunk(u32 &id);

	/// Examines header and size of next chunk (no start of new chunk).
	/// Used to determine which chunk is coming next.
	/// \param id Identifier of next chunk
	/// \return Size of next chunk, 0 if end of file (EOF)
	u32 ExamineNextChunk(u32 &id);

	inline void Read(void* data, size_t count) { if(f && !writing) f->Read(data, count); }
	inline void Write(void* data, size_t count) { if(f && writing) f->Write(data, count); }

	inline u8 ReadU8() { if(f && !writing) return f->ReadU8(); else return 0; }
	inline u16 ReadU16() { if(f && !writing) return f->ReadU16(); else return 0; }
	inline u32 ReadU32() { if(f && !writing) return f->ReadU32(); else return 0; }
	inline float ReadFloat() { if(f && !writing) return f->ReadFloat(); else return 0; }

	inline void WriteU8(u8 data) { Write(&data, sizeof(u8)); }
	inline void WriteU16(u16 data) { Write(&data, sizeof(u16)); }
	inline void WriteU32(u32 data) { Write(&data, sizeof(u32)); }
	inline void WriteFloat(float data) { Write(&data, sizeof(float)); }

/*	inline void readArrayU16(u16 *data, size_t count)
	{
		if(f && !writing) f->readArrayU16(data, count);
	}
	inline void readArrayU32(u32 *data, size_t count)
	{
		if(f && !writing) f->readArrayU32(data, count);
	}
	inline void readArrayFloat(float *data, size_t count)
	{
		if(f && !writing) f->readArrayFloat(data, count);
	}*/
};


