#pragma once
#include <math.h>

inline float radToDeg( float rad ) { return (float) ( (rad * 180.0f) / 3.1415f );	}
inline float degToRad( float degree ) { return (float) (( degree *  3.1415f)/180.0f ); }

//-----------------------------------------------------------------
// Vec2f
//-----------------------------------------------------------------
class Vec2f
{
public:
	float x, y;

	Vec2f() {}
	Vec2f(float fx, float fy) : x(fx), y(fy) {}
};

//-----------------------------------------------------------------
// Vec3
//-----------------------------------------------------------------
class Vec3f
{
public:
	float x, y, z;

	Vec3f() {}
	Vec3f(float fx, float fy, float fz) : x(fx), y(fy), z(fz) {}
	inline void SetXYZ(float fx, float fy, float fz) { x=fx; y=fy; z=fz; };

	Vec3f& operator += (const Vec3f& v) { x += v.x; y += v.y; z += v.z; return *this; }
	Vec3f& operator -= (const Vec3f& v) { x -= v.x; y -= v.y; z -= v.z; return *this; }
	Vec3f& operator *= (float f) { x *= f; y *= f; z *= f; return *this; }
	Vec3f operator + (const Vec3f& v) const { return Vec3f(x + v.x, y + v.y, z + v.z); }
	Vec3f operator - (const Vec3f& v) const { return Vec3f(x - v.x, y - v.y, z - v.z); }
	friend Vec3f operator * (Vec3f& v, float f) { return Vec3f(v.x * f, v.y * f, v.z * f); }
	Vec3f operator - () const { return Vec3f(-x, -y, -z); }
	bool operator==(const Vec3f& v) const { return (v.x == x) && (v.y == y) && (v.z == z); }

	inline float LengthSquared() { return x*x+y*y+z*z; }
	inline float Length() { return (float)sqrtf(x*x+y*y+z*z); }

	inline void Normalize() {
		float s = Length();
		if (s) s = 1.0f/s;
		x*=s; y*=s; z*=s;
	}
};

inline Vec3f operator / (Vec3f v1, float s) { return v1*(1.0f/s); }

inline float DotProduct(const Vec3f& a, const Vec3f& b) { return a.x*b.x+a.y*b.y+a.z*b.z; }

inline Vec3f CrossProduct(const Vec3f& a, const Vec3f& b) {
	return Vec3f((a.y*b.z)-(a.z*b.y),(a.z*b.x)-(a.x*b.z),(a.x*b.y)-(a.y*b.x));
}

inline Vec3f Normalize(const Vec3f& v) {
	float s=(float)sqrtf(v.x*v.x+v.y*v.y+v.z*v.z);
	if(s) s=1/s;
	return Vec3f(v.x*s,v.y*s,v.z*s);
}

//-----------------------------------------------------------------
// Vec4f
//-----------------------------------------------------------------
class Vec4f
{
public:
	float x, y, z, w;

	Vec4f() {}
	Vec4f(float fx, float fy, float fz, float fw) : x(fx), y(fy), z(fz), w(fw) {}
	Vec4f(const Vec3f& v3, float vw) : x(v3.x), y(v3.y), z(v3.z), w(vw) {}
	inline void SetXYZW(float fx, float fy, float fz, float fw) { x=fx; y=fy; z=fz; w=fw; };
};

inline float DotProduct(const Vec4f& a, const Vec4f& b) { return a.x*b.x+a.y*b.y+a.z*b.z+a.w*b.w; }

//-----------------------------------------------------------------
// Quat
//-----------------------------------------------------------------
class Mtx4f;

class Quat
{
public:
	float x, y, z, w;
	Quat() { x=y=z=0.0f; w=1.0f; };
	Quat(float fx, float fy, float fz, float fw) : x(fx), y(fy), z(fz), w(fw) {}
	Quat(Vec3f v, float t) { v.Normalize(); w=(float)cos(t/2.0); v=v*(float)sin(t/2.0); x=v.x; y=v.y; z=v.z; };
	inline float Angle(){return (float)(acos(w)*2.0);}
	inline Vec3f Axis() { Vec3f a(x,y,z); return a*(float)(1/sin(Angle()/2.0));}
	inline Vec3f xdir() { return Vec3f(1-2*(y*y+z*z),  2*(x*y+w*z),  2*(x*z-w*y));}
	inline Vec3f ydir() { return Vec3f(  2*(x*y-w*z),1-2*(x*x+z*z),  2*(y*z+w*x));}
	inline Vec3f zdir() { return Vec3f(  2*(x*z+w*y),  2*(y*z-w*x),1-2*(x*x+y*y));}

	Quat operator*(const float s) const { return Quat(x*s, y*s, z*s, w*s); }

	inline Quat operator+(const Quat& q) { return Quat(x + q.x, y + q.y, z + q.z, w + q.w); }
    inline Quat operator-(const Quat& q) { return Quat(x - q.x, y - q.y, z - q.z, w - q.w); }

	inline void ComputeW() {
		float t = 1.0f - (x*x) - (y*y) - (z*z);
		if (t < 0.0f) w = 0.0f;
		else w = -sqrtf(t);
	}
	inline void Normalize()
	{
		float mag = sqrtf(x*x + y*y + z*z + w*w);

		if (mag > 0.0f) {
			float oneOverMag = 1.0f / mag;

			x *= oneOverMag;
			y *= oneOverMag;
			z *= oneOverMag;
			w *= oneOverMag;
		}
	}

	// Multiply quaternion by vector
	Quat operator *(const Vec3f& v)
	{
		Quat ret;
		ret.w = - (x * v.x) - (y * v.y) - (z * v.z);
		ret.x =   (w * v.x) + (y * v.z) - (z * v.y);
		ret.y =   (w * v.y) + (z * v.x) - (x * v.z);
		ret.z =   (w * v.z) + (x * v.y) - (y * v.x);
		return ret;
	}

	Vec3f RotateVector(const Vec3f &v);
	void Slerp(const Quat& from, const Quat& to, float t);
};

Quat operator*(Quat a, Quat b);

//-----------------------------------------------------------------
// Mtx4f
//-----------------------------------------------------------------
class Mtx4f
{
public:
	Vec4f m[4];	// columns (GL)

	Mtx4f() {}
	Mtx4f(Vec4f x, Vec4f y, Vec4f z, Vec4f w) { m[0]=x; m[1]=y; m[2]=z; m[3]=w; }
	inline void Set(float m00, float m10, float m20, float m30,		// <-- attention: transposed layout!
					float m01, float m11, float m21, float m31,
					float m02, float m12, float m22, float m32,
					float m03, float m13, float m23, float m33) {
		m[0].SetXYZW(m00, m01, m02, m03);
		m[1].SetXYZW(m10, m11, m12, m13);
		m[2].SetXYZW(m20, m21, m22, m23);
		m[3].SetXYZW(m30, m31, m32, m33);
	}

	void FromQuat(Quat q);

	inline void Identity() { 
		m[0].SetXYZW(1, 0, 0, 0);
		m[1].SetXYZW(0, 1, 0, 0);
		m[2].SetXYZW(0, 0, 1, 0);
		m[3].SetXYZW(0, 0, 0, 1);
	}

	static Mtx4f TranslationMatrix(const Vec3f& t);
	static Mtx4f RotationMatrixX(float angle);
	static Mtx4f RotationMatrixY(float angle);
	static Mtx4f RotationMatrixZ(float angle);

	operator const float* () const { return &m[0].x; }

	inline void SetScale(const Vec3f& s) { m[0].x=s.x; m[1].y=s.y; m[2].z=s.z; }
	inline Vec3f GetScale() const { return Vec3f(m[0].x,m[1].y,m[2].z); }

	inline void SetTranslation(const Vec3f& t) { m[3].x=t.x; m[3].y=t.y; m[3].z=t.z; }
	inline Vec3f GetTranslation() const { return Vec3f(m[3].x, m[3].y, m[3].z); }
	void Translate(const Vec3f& t);

	// get upper 3x3 rotation part as 4x4 matrix
	Mtx4f GetOrientation();

	inline void SetRow0(const Vec4f& v) { m[0].x=v.x; m[1].x=v.y; m[2].x=v.z; m[3].x=v.w; }
	inline void SetRow1(const Vec4f& v) { m[0].y=v.x; m[1].y=v.y; m[2].y=v.z; m[3].y=v.w; }
	inline void SetRow2(const Vec4f& v) { m[0].z=v.x; m[1].z=v.y; m[2].z=v.z; m[3].z=v.w; }
	inline void SetRow3(const Vec4f& v) { m[0].w=v.x; m[1].w=v.y; m[2].w=v.z; m[3].w=v.w; }
	inline Vec4f GetRow0() const { return Vec4f(m[0].x, m[1].x, m[2].x, m[3].x); }	// right
	inline Vec4f GetRow1() const { return Vec4f(m[0].y, m[1].y, m[2].y, m[3].y); }	// up
	inline Vec4f GetRow2() const { return Vec4f(m[0].z, m[1].z, m[2].z, m[3].z); }	// forward
	inline Vec4f GetRow3() const { return Vec4f(m[0].w, m[1].w, m[2].w, m[3].w); }
	inline Mtx4f GetTranspose() const { return Mtx4f(GetRow0(),GetRow1(),GetRow2(),GetRow3()); }
	inline void Transpose() { Mtx4f mat = GetTranspose(); *this=mat; }

	/*inline void Translation(const Vec3f& t) {
		Identity();
		SetTranslation(t);
	}*/
	inline Mtx4f operator*(Mtx4f m2) const {
		Mtx4f res;
		for (int col=0; col < 4; col++) {
			res.m[col] = Vec4f(	DotProduct(GetRow0(), m2.m[col]),
								DotProduct(GetRow1(), m2.m[col]),
								DotProduct(GetRow2(), m2.m[col]),
								DotProduct(GetRow3(), m2.m[col]));
		}
		return res;
	}

	void Invert();
	Mtx4f GetInverted() const;

	void LookAt(const Vec3f& pos, const Vec3f& target, const Vec3f& up);
	// TODO: fromQuat
};



typedef Vec2f Vec2;
typedef Vec3f Vec3;
typedef Vec4f Vec4;

class Projection {
public:
	float fov;
	float aspect;
	float zNear;
	float zFar;

	inline void Set(float Fov, float Aspect, float ZNear, float ZFar) {
		fov = Fov; aspect = Aspect; zNear = ZNear; zFar = ZFar;
		UpdateMatrix();
	}
	//inline void SetPerspective(float fov, float aspect, float zNear, float zFar) {
	inline void UpdateMatrix() {
		float top = zNear * tan(0.5f * fov * 3.1415f / 180.0f);
		float right	= top * aspect;
		float bottom = -top;
		float left = -right;
		SetFrustum(left, right, top, bottom, zNear, zFar);
	}
	inline const Mtx4f& GetMatrix() const { return matrix; }
protected:
	inline void SetFrustum(float left, float right, float top, float bottom, float zNear, float zFar) {
		matrix.m[0].SetXYZW(2*zNear/(right-left), 0, 0, 0);
		matrix.m[1].SetXYZW(0, 2*zNear/(top-bottom), 0, 0);
		matrix.m[2].SetXYZW((right+left)/(right-left), (top+bottom)/(top-bottom), -(zFar+zNear)/(zFar-zNear), -1);
		matrix.m[3].SetXYZW(0, 0, -2*zFar*zNear/(zFar-zNear), 0);
	}

	Mtx4f matrix;
};

class BBox {
public:
	Vec3f vmin;
	Vec3f vmax;
};

